package main

import (
	"io"
	"log"
	"os"
	"sync"
	"time"

	"www.bamsoftware.com/git/turbotunnel-paper.git/example/turbotunnel/turbotunnel"

	"github.com/xtaci/kcp-go/v5"
	"github.com/xtaci/smux"
)

func main() {
	if len(os.Args) != 2 {
		log.Fatalf("usage: %s IP:PORT", os.Args[0])
	}
	addr := os.Args[1]
	if err := run(addr); err != nil {
		log.Fatal(err)
	}
}

func run(addr string) error {
	// Generate a new SessionID to identify this session.
	sessionID := turbotunnel.NewSessionID()

	// As the packet-sending and -receiving interface, use a sequence of
	// redialed TCP connections.
	pconn := NewRedialPacketConn(sessionID, addr)

	// The KCP net.Conn is a reliable channel over pconn.
	conn, err := kcp.NewConn2(pconn.RemoteAddr(), nil, 0, 0, pconn)
	if err != nil {
		return err
	}
	defer conn.Close()
	log.Printf("session %v: begin", sessionID)
	defer log.Printf("session %v: end", sessionID)

	// Some parameter tweaks to improve performance.
	// Permit coalescing the payloads of consecutive sends.
	conn.SetStreamMode(true)
	// Disable the dynamic congestion window (limit only by the maximum of
	// local and remote static windows).
	conn.SetNoDelay(
		0, // default nodelay
		0, // default interval
		0, // default resend
		1, // nc=1 => congestion window off
	)
	// Increase the (send, receive) window sizes from the default of
	// (32, 32).
	conn.SetWindowSize(1024, 1024)

	// Add smux stream multiplexing and timeout over the KCP net.Conn.
	smuxConfig := smux.DefaultConfig()
	smuxConfig.Version = 2
	smuxConfig.KeepAliveTimeout = 1 * time.Minute
	smuxConfig.MaxReceiveBuffer = 4 * 1024 * 1024 // default is 4 * 1024 * 1024
	smuxConfig.MaxStreamBuffer = 1 * 1024 * 1024  // default is 65536
	sess, err := smux.Client(conn, smuxConfig)
	if err != nil {
		return err
	}
	defer sess.Close()

	// Open a smux stream.
	stream, err := sess.OpenStream()
	if err != nil {
		return err
	}
	log.Printf("session %v stream %v: begin", sessionID, stream.ID())
	defer log.Printf("session %v stream %v: end", sessionID, stream.ID())

	handleStream(stream)
	return nil
}

// Read from the stream and write to stdout; read from stdin and write to the
// stream.
func handleStream(stream *smux.Stream) {
	var wg sync.WaitGroup
	wg.Add(2)
	go func() {
		_, err := io.Copy(os.Stdout, stream)
		if err != nil {
			log.Printf("%v: error copying to stdout: %v", stream.RemoteAddr(), err)
		}
		os.Stdout.Close()
		wg.Done()
	}()
	go func() {
		_, err := io.Copy(stream, os.Stdin)
		if err != nil {
			log.Printf("%v: error copying from stdin: %v", stream.RemoteAddr(), err)
		}
		stream.Close()
		wg.Done()
	}()
	wg.Wait()
}
