package main

import (
	"io"
	"log"
	"net"
	"os"
	"time"

	"github.com/xtaci/kcp-go/v5"
	"github.com/xtaci/smux"
)

func main() {
	if len(os.Args) != 2 {
		log.Fatalf("usage: %s IP:PORT", os.Args[0])
	}
	addr := os.Args[1]
	if err := run(addr); err != nil {
		log.Fatal(err)
	}
}

func run(addr string) error {
	// This listener will accept the TCP connections that carry encapsulated
	// packets.
	tcpLn, err := net.Listen("tcp", addr)
	if err != nil {
		return err
	}
	defer tcpLn.Close()

	// Build a net.PacketConn interface around the connections accepted by
	// the TCP listener.
	pconn := NewListenerPacketConn(tcpLn)

	// Create a listener for the virtual KCP sessions that exist on top of
	// the physical TCP sessions.
	ln, err := kcp.ServeConn(nil, 0, 0, pconn)
	if err != nil {
		return err
	}

	return acceptSessions(ln)
}

func acceptSessions(ln *kcp.Listener) error {
	for {
		conn, err := ln.AcceptKCP()
		if err != nil {
			if err, ok := err.(net.Error); ok && err.Temporary() {
				continue
			}
			return err
		}

		// Some parameter tweaks to improve performance.
		// Permit coalescing the payloads of consecutive sends.
		conn.SetStreamMode(true)
		// Disable the dynamic congestion window (limit only by the
		// maximum of local and remote static windows).
		conn.SetNoDelay(
			0, // default nodelay
			0, // default interval
			0, // default resend
			1, // nc=1 => congestion window off
		)
		// Increase the (send, receive) window sizes from the default of
		// (32, 32).
		conn.SetWindowSize(1024, 1024)

		go func() {
			defer conn.Close()
			log.Printf("session %v: begin", conn.RemoteAddr())
			defer log.Printf("session %v: end", conn.RemoteAddr())
			err := acceptStreams(conn)
			if err != nil {
				log.Printf("session %v: acceptStreams: %v", conn.RemoteAddr(), err)
			}
		}()
	}
}

func acceptStreams(conn *kcp.UDPSession) error {
	// Add smux stream multiplexing and timeout over the KCP net.Conn.
	smuxConfig := smux.DefaultConfig()
	smuxConfig.Version = 2
	smuxConfig.KeepAliveTimeout = 1 * time.Minute
	smuxConfig.MaxReceiveBuffer = 4 * 1024 * 1024 // default is 4 * 1024 * 1024
	smuxConfig.MaxStreamBuffer = 1 * 1024 * 1024  // default is 65536
	sess, err := smux.Server(conn, smuxConfig)
	if err != nil {
		return err
	}
	defer sess.Close()

	for {
		stream, err := sess.AcceptStream()
		if err != nil {
			if err, ok := err.(net.Error); ok && err.Temporary() {
				continue
			}
			return err
		}
		go func() {
			defer stream.Close()
			log.Printf("session %v stream %v: begin", conn.RemoteAddr(), stream.ID())
			defer log.Printf("session %v stream %v: end", conn.RemoteAddr(), stream.ID())
			err := handleStream(stream)
			if err != nil {
				log.Printf("session %v stream %v: handleStream: %v", conn.RemoteAddr(), stream.ID(), err)
			}
		}()
	}
}

func handleStream(stream *smux.Stream) error {
	n, err := io.Copy(stream, stream)
	log.Printf("session %v stream %v: echoed %v bytes", stream.RemoteAddr(), stream.ID(), n)
	if err == io.EOF {
		err = nil
	}
	return err
}
